"""Pydantic classes for validating parameters and configurations for `Chinchilla` and `Simulator`."""

from __future__ import annotations  # PEP 604 backport

from pydantic import BaseModel, Field, model_validator, field_validator

from typing import Optional, Callable, Iterable, Iterator, Tuple, Dict
# We use capitalized types here because built-in generic types (PEP 585) supported only from python 3.9

VALID_KEYS_IN_LOWER = "e", "a", "b", "alpha", "beta"


class ParamGrid(BaseModel):
    """
    Validates a grid of initialization for scaling law (/loss predictor) parameters.

    Attributes:
        E or e: Tuple of floats representing initial values for the E parameter or its log form.
        A or a: Tuple of floats representing initial values for the A parameter or its log form.
        B or b: Tuple of floats representing initial values for the B parameter or its log form.
        alpha: Tuple of floats representing initial values for the alpha parameter.
        beta: Tuple of floats representing initial values for the beta parameter.
    """

    E: Optional[Tuple[float, ...]] = None
    A: Optional[Tuple[float, ...]] = None
    B: Optional[Tuple[float, ...]] = None
    alpha: Tuple[float, ...]
    beta: Tuple[float, ...]
    e: Optional[Tuple[float, ...]] = None
    a: Optional[Tuple[float, ...]] = None
    b: Optional[Tuple[float, ...]] = None

    @model_validator(mode="before")
    def check_keys(cls, values):
        """Validates that the parameter grid contains the correct keys."""
        keys = [k.lower() for k in values.keys()]
        if len(set(keys)) != 5:
            raise ValueError("`param_grid` must be a dict of 5 keys")
        if len(set(keys)) != len(keys):
            raise ValueError("There are effectively duplicate keys when lowercased.")

        invalid_keys = [
            k
            for k in values.keys()
            if (len(k) == 1 and k.lower() not in VALID_KEYS_IN_LOWER) or (len(k) > 1 and k not in VALID_KEYS_IN_LOWER)
        ]
        missing_keys = [p for p in VALID_KEYS_IN_LOWER if p not in [k.lower() for k in keys]]
        missing_keys = [f"{p.upper()}/{p}" if len(p) == 1 else p for p in missing_keys]
        # duplicate_keys = []
        if invalid_keys + missing_keys:
            raise ValueError(
                "Invalid set of keys in `param_grid`:\n"
                + (f"- Invalid key(s): {', '.join(invalid_keys)}\n" if invalid_keys else "")
                + (f"- Missing key(s): {', '.join(missing_keys)}\n" if missing_keys else "")
            )
        return values


class SeedRanges(BaseModel):
    """
    Validates the regime of seed models.

    Attributes:
        C: Tuple of two floats representing the range for the `C` parameter.
        N_to_D: Tuple of two floats representing the range for the `N_to_D` parameter.
    """

    C: Tuple[float, float]
    N_to_D: Tuple[float, float]

    @field_validator("C")
    def check_C_values(cls, values):
        """Validates the values for the 'C' range."""
        if any(c < 0 for c in values):
            raise ValueError("Negative value(s) in 'C': Consider changing data type.")
        if not values[0] < values[1]:
            raise ValueError("'C' must have the first value less than the second value.")
        return values


class ModelSearchConfig(BaseModel):
    """
    Validates keyword arguments into `_utils.search_model_config`.

    Attributes:
        size_estimator: A callable that estimates the size of the model.
        hyperparam_grid: A dictionary representing the hyperparameter grid.
    """

    size_estimator: Callable[[], float]
    hyperparam_grid: Dict[str, Tuple[float, ...]]

    @field_validator("size_estimator")
    def size_estimator_must_be_callable(cls, v):
        """Validates that the size estimator is a callable."""
        if not callable(v):
            raise TypeError("size_estimator must be a callable (function or class method).")
        return v

    @field_validator("hyperparam_grid")
    def hyperparam_grid_must_be_dict(cls, v):
        """Validates that the hyperparameter grid is a dictionary."""
        if not isinstance(v, dict):
            raise TypeError("hyperparam_grid must be a dictionary.")
        return v


class SimulationArgs(BaseModel):
    """
    Validates arguments for running simulations.

    Attributes:
        num_seeding_steps: The number of seeding steps.
        num_scaling_steps: The number of scaling steps.
        target_params: A dictionary representing hypothetical scaling law parameters.
        noise_generator: An optional iterable of floats representing the additional loss randomly caused by imperfect training.
        scaling_factor: An optional float representing the scaling factor in FLOPs.
    """

    num_seeding_steps: int = Field(gt=0)
    num_scaling_steps: int = Field(ge=0)
    target_params: Dict[str, float] = (
        {"E": 1.69337368, "A": 406.401018, "B": 410.722827, "alpha": 0.33917084, "beta": 0.2849083},
    )
    noise_generator: Optional[Iterable[float]] = None
    scaling_factor: Optional[float] = None

    @model_validator(mode="before")
    def check_seeding_steps(cls, values):
        """Validates the number of seeding and scaling steps along with the scaling factor."""
        num_seeding_steps = values.get("num_seeding_steps")
        num_scaling_steps = values.get("num_scaling_steps")
        scaling_factor = values.get("scaling_factor")  # Updated with self.scaling_factor if None
        if num_seeding_steps < 3 and num_scaling_steps:
            raise ValueError("Too few seed models to estimate the scaling law.")
        if num_scaling_steps and scaling_factor is None:
            raise ValueError(
                "Chinchilla object must be assigned a positive float to `scaling_factor` when simulating the scaling process. "
                "You may specify it in `cc = Chinchilla(...)` or `cc.simulate(...)`"
            )
        return values

    @field_validator("target_params")
    def check_target_params(cls, v):
        """Validates that the target parameters are a dictionary with float or int values."""
        if not isinstance(v, dict) or not all(isinstance(p, (float, int)) for p in v.values()):
            raise ValueError("target_params must be a dictionary with float or int values.")
        return v

    @field_validator("noise_generator")
    def check_noise_generator(cls, v):
        """Validates the noise generator, ensuring it is an iterator or None."""
        if v is not None and not isinstance(v, Iterator):
            raise ValueError("noise_generator must be an iterator or None")
        return v



from __future__ import annotations  # PEP 604 backport

import os

import pandas as pd

from ._logger import get_logger


class Database:
    """
    Stores and manipulates scaling data in a Pandas DataFrame the default persistence to a CSV file.
    The Database class is used internally by a `Chinchilla` instance.

    If `project_dir` is provided, the DataFrame is initialized from the CSV file at that location.
    If the file does not exist or is empty, a new DataFrame is created. If `project_dir` is None,
    the DataFrame is kept in memory.

    **Default columns**:
    - `C` (float): Compute in FLOPs.
    - `N` (int): Number of parameters.
    - `D` (int): Number of data samples seen.
    - `loss` (float): Loss value (optional, use case dependent).

    Args:
        project_dir (str | None): Directory for the CSV file storage.
        columns (list[str]): Column names for the DataFrame.
        logger (Logger): Logger instance for database messages.
    """

    def __init__(
        self, project_dir: str | None = None, columns: list[str] = ["C", "N", "D", "loss"], log_level: int = 30
    ) -> None:
        """
        Initializes the Database instance.

        Args:
            project_dir (Optional[str]): The directory path to save the DataFrame as a CSV file.
                                         If None, the DataFrame will not be saved to disk.
            columns (List[str]): A list of column names for the DataFrame.
            log_level (int): The logging level for the logger instance.
        """
        self.logger = get_logger(log_level, name="chinchilla.database")
        self.project_dir = project_dir
        self.filepath = os.path.join(self.project_dir, "df.csv") if self.project_dir else None

        if self.filepath and (os.path.exists(self.filepath) and os.path.getsize(self.filepath) > 0):
            self.df = pd.read_csv(self.filepath)
            self.logger.debug("Data types:\n" + self.df.dtypes.T.to_string())
        else:
            self.logger.info(
                f"Initializing a database to be saved to: {self.filepath}"
                if self.filepath
                else "Initializing an in-memory database for Simulator."
            )
            # We define an empty database for when referencing the number of existing data points
            self.df = pd.DataFrame([], columns=columns)

    def append(self, **result: dict[str, float]) -> None:
        """
        Appends a new row of results to the DataFrame and updates the CSV file if `project_dir` is set.

        If 'C' is not provided in `result`, it is automatically calculated as $6ND$.
        All numerical values are rounded to the nearest integer to prevent scientific notation in large values.
        Additional columns provided by the user are appended to the DataFrame.

        Args:
            result (dict[str, float]): A dictionary containing the data to append. Must include 'N', 'D', and 'loss' keys.
                If 'C' is not provided in `result`, it is automatically calculated as $6ND$.
                All numerical values are rounded to the nearest integer to prevent losing precisions to scientific notation for large values.
                Additional columns provided by the user will be appended to the DataFrame without any conflicts.
        """
        if not result or {"N", "D", "loss"} - set(result.keys()):
            raise ValueError("The 'result' dictionary must contain 'N', 'D', and 'loss' keys.")

        if "C" not in result:
            result["C"] = 6 * result["N"] * result["D"]
        for k in ["C", "N", "D"]:
            result[k] = round(result[k])  # This helps prevent scientific notation of large values

        # Collect all columns added by the user
        cols_additional = [c for c in result.keys() if c not in self.df.columns]
        record = pd.DataFrame([result], columns=self.df.columns.tolist() + cols_additional)

        # self.df = pd.concat([self.df, record], ignore_index=True); raise NotImplementedError("Do not use this code.")
        # Concatenating to the empty database may cause a data type issue:
        # e.g., when calling `Chinchilla.predict_loss` at the end of `Chinchilla.fit` (fixed with `.astype(float)`)
        self.df = record if len(self.df) == 0 else pd.concat([self.df, record], ignore_index=True)

        if self.filepath:
            self.df.to_csv(self.filepath, index=False, float_format="%f")
